import duckdb
import time
import traffic_anomaly
from .data_loader import load_data
from .data_aggregator import aggregate_data
from .data_saver import save_data
from .utils import round_down_15
from .utils import v_print
import os


class SignalDataProcessor:
    '''
    Main class in the atspm package, used to process signal data by turning raw hi-res data into aggregated data.

    This class handles the entire pipeline of loading raw data, performing various aggregations,
    and saving the results. It supports both one-time processing and incremental processing with
    unmatched event handling. Most the inputs here are optional, depending on the desired processing.

    Attributes
    ----------
    raw_data : str or DataFrame
        The raw data to be processed, either as a file path or a DataFrame.
    detector_config : str or DataFrame
        The detector configuration, either as a file path or a DataFrame.
    bin_size : int
        The size of the time bins for aggregation, in minutes.
    output_dir : str
        The directory where the output files will be saved.
    output_to_separate_folders : bool
        If True, output files will be saved in separate folders.
    output_format : str
        The format of the output files. Options are "csv", "parquet", "json".
    output_file_prefix : str
        Prefix to be added to all output file names.
    remove_incomplete : bool
        If True, removes periods with incomplete data based on the 'has_data' aggregation.
    unmatched_event_settings : dict, optional
        Settings for handling unmatched events in incremental processing. Includes:
        - df_or_path: str, path to save/load unmatched events
        - split_fail_df_or_path: str, path to save/load unmatched split failure events
        - max_days_old: int, maximum age of unmatched events to consider
    to_sql : bool
        If True, returns SQL strings instead of executing queries.
    verbose : int
        Controls the verbosity of output. 0: only errors, 1: performance, 2: debug statements.
    aggregations : list of dict
        A list of dictionaries, each containing the name of an aggregation function and its parameters.
        Supported aggregations include: 'has_data', 'actuations', 'arrival_on_green', 'communications',
        'coordination', 'ped', 'unique_ped', 'full_ped', 'split_failures', 'splits', 'terminations',
        'yellow_red', 'timeline', and potentially others.

    Methods
    -------
    load()
        Loads the raw data and detector configuration into DuckDB tables.
    aggregate()
        Runs all specified aggregations on the loaded data.
    save()
        Saves the processed data to the specified output directory and format.
    close()
        Closes the database connection.
    run()
        Executes the complete data processing pipeline: load, aggregate, save, and close.
        If to_sql is True, returns the SQL queries instead of executing them.

    Example
    -------
    processor = SignalDataProcessor(
        raw_data=sample_data.data,
        detector_config=sample_data.config,
        bin_size=15,
        output_dir='test_folder',
        output_to_separate_folders=True,
        output_format='csv',
        remove_incomplete=True,
        unmatched_event_settings={
            'df_or_path': 'test_folder/unmatched.parquet',
            'split_fail_df_or_path': 'test_folder/sf_unmatched.parquet',
            'max_days_old': 14
        },
        verbose=1,
        aggregations=[
            {'name': 'has_data', 'params': {'no_data_min': 5, 'min_data_points': 3}},
            {'name': 'actuations', 'params': {}},
            # ... other aggregations ...
        ]
    )
    processor.run()
    '''

    def __init__(self, **kwargs):
        """Initializes the SignalDataProcessor with the provided keyword arguments."""
        # Optional parameters
        self.raw_data = None
        self.detector_config = None
        self.unmatched_event_settings = None # For incremental processing of timeline, split failure, arrival on green, and yellow red)
        self.unmatched_found = False
        self.incremental_run = False
        self.binned_actuations = None # For detector_health aggregation
        self.device_groups = None # For detector_health aggregation if groups are provided
        self.remove_incomplete = False
        self.to_sql = False
        self.verbose = 1 # 0: only print errors, 1: print performance, 2: print debug statements
        
        # Extract parameters from kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

        # Check for valid bin_size and no_data_min combo
        if self.remove_incomplete:
            # raise error if 'has_data' aggregation is not in aggregations
            assert any(d['name'] == 'has_data' for d in self.aggregations), "Remove_incomplete requires 'has_data' aggregation!"
            # extract has_data parameters
            no_data_min = next(x['params']['no_data_min'] for x in self.aggregations if x['name'] == 'has_data')
            assert self.bin_size % no_data_min == 0, "bin_size / no_data_min must be a whole number"
            # Make sure that has_data is the first aggregation
            idx = [d['name'] for d in self.aggregations].index('has_data')
            self.aggregations.insert(0, self.aggregations.pop(idx))

        # Check format of unmatched_event_settings
        # Duckdb needs quotes if it is a file path, but not if it is a dataframe
        if self.unmatched_event_settings is not None:
            self.incremental_run = True
            self.unmatched_found = True
            for key, value in self.unmatched_event_settings.items():
                if key == 'max_days_old':
                    continue
                if isinstance(value, str):
                    if os.path.exists(value):
                        self.unmatched_event_settings[key] = f"'{value}'"
                    else:
                        v_print(f"Warning! unmatched_events file {value} does not exist at given path yet. This is expected if no unmatched events have been recorded (first run), and a file will be output to that path at end of this run.", self.verbose)
                        self.unmatched_found = False

        # Check if detector_health is in aggregations
        if any(d['name'] == 'detector_health' for d in self.aggregations):
            try:
                idx = [d['name'] for d in self.aggregations].index('detector_health')
                self.binned_actuations = self.aggregations[idx]['params']['data']
                self.device_groups = self.aggregations[idx]['params']['device_groups']
            except KeyError:
                raise ValueError("detector_health aggregation requires 'data' and 'device_groups' parameters. 'device_groups' can be set to None.")
     
        # Establish a connection to the database
        self.conn = duckdb.connect()
        # Track whether data has been loaded
        self.data_loaded = False

        # Use connection to get current timestamp
        # This is a placeholder for when to_sql is True. After the class is instantiated, 
        # timestamps need to be set by the user for the full_ped query to work.
        self.max_timestamp = self.conn.execute("SELECT CURRENT_TIMESTAMP").fetchone()[0]
        self.min_timestamp = self.max_timestamp

    def load(self):
        """Loads raw data and detector configuration into DuckDB tables."""
        if self.data_loaded:
            v_print("Data already loaded! Reinstantiate the class to reload data.", self.verbose)
            return
        if self.to_sql:
            v_print("to_sql option is True, data will not be loaded.", self.verbose)
            return
        try:
            load_data(self.conn,
                    self.verbose,
                    self.raw_data,
                    self.detector_config,
                    self.unmatched_event_settings,
                    self.unmatched_found)
            # delete self.raw_data and self.detector_config to free up memory
            self.data_loaded = True
            if self.raw_data is not None:
                self.min_timestamp = self.conn.execute("SELECT MIN(timestamp) FROM raw_data").fetchone()[0]
                self.max_timestamp = self.conn.execute("SELECT MAX(timestamp) FROM raw_data").fetchone()[0]
                v_print(f'Data loaded from {self.min_timestamp} to {self.max_timestamp}', self.verbose)
            # free up memory
            del self.raw_data
            del self.detector_config
        except Exception as e:
            v_print('*'*50, self.verbose)
            v_print('WARNING: problem loading data!', self.verbose)
            v_print('Make sure raw_data column names are: TimeStamp, DeviceId, EventId, Parameter', self.verbose)
            v_print('Make sure detector_config column names are: DeviceId, Phase, Parameter, Function', self.verbose)
            v_print('*'*50, self.verbose)
            raise e
        
    def aggregate(self):
        """Runs all aggregations."""
        # Instantiate a dictionary to store runtimes
        self.runtimes = {}
        self.sql_queries = {} # for storing sql string when to_sql is True

        # Create unmatched_events table if unmatched_events is not None
        # This table will be used to insert unmatched events, to be saved and reloaded in the next run
        #if self.unmatched_event_settings is not None:
        #    v_print("Creating unmatched_events table", self.verbose, level=2)
        #    self.conn.query(f"CREATE TABLE unmatched_events AS SELECT * AS aggregation FROM raw_data WHERE 1=0")

        for aggregation in self.aggregations:
            start_time = time.time()

            #######################
            ### Detector Health ###
            ### Does Not Use aggregate_data function
            ### Relies on traffic-anomaly package instead
            # Decompose data
            if aggregation['name'] == 'detector_health':
                if self.to_sql:
                    raise ValueError("to_sql option is  supported for detector_health")
                decomp = traffic_anomaly.median_decompose(
                    self.binned_actuations,
                    **aggregation['params']['decompose_params']
                )
                del self.binned_actuations
                # Join groups to decomp
                if self.device_groups is not None:
                    device_groups = self.device_groups # DuckDB needs a direct pointer to see the table
                    decomp = self.conn.sql("SELECT * FROM decomp NATURAL JOIN device_groups").df()
                    # Exclude group_grouping_columns in anomaly_params
                    exclude_col = ', '.join(["'{}'".format(x) for x in aggregation['params']['anomaly_params']['group_grouping_columns']])
                    exclude_col = f"EXCLUDE ({exclude_col})"
                    
                else:
                    exclude_col = ""
                # Find Anomalies
                anomaly = traffic_anomaly.find_anomaly(
                    decomposed_data=decomp,
                    **aggregation['params']['anomaly_params']
                )
                # Extract max date from anomaly table and subtract return_last_n_days
                sql = f"""
                    SELECT CAST(MAX(TimeStamp)::DATE - INTERVAL '{aggregation['params']['return_last_n_days']-1}' DAY AS VARCHAR) AS max_date_minus_one
                    FROM anomaly
                    """
                max_date = self.conn.query(sql).fetchone()[0]

                # Save anomaly table to DuckDB
                query = f"""CREATE OR REPLACE TABLE detector_health AS
                        SELECT * {exclude_col}
                        FROM anomaly
                        WHERE TimeStamp >= '{max_date}'
                        """
                self.conn.execute(query)
                # no external sql file like other aggregations, so just continue
                end_time = time.time()
                self.runtimes[aggregation['name']] = end_time - start_time
                continue

            # Add bin_size and remove_incomplete to params
            aggregation['params']['bin_size'] = self.bin_size
            aggregation['params']['remove_incomplete'] = self.remove_incomplete
            aggregation['params']['from_table'] = 'raw_data'

            #######################
            ### Full Pedestrian ###
            # Add min_timestamp and max_timestamp to params if detector_faults or full_ped
            if aggregation['name'] == 'full_ped':
                # Round min_timestamp down to nearest bin_size
                aggregation['params']['min_timestamp'] = round_down_15(self.min_timestamp)
                aggregation['params']['max_timestamp'] = round_down_15(self.max_timestamp)

            #######################
            ### Unmatched Events ##
            # If unmatched_event_settings is supplied, then change the from_table for timeline, split_failures, arrival_on_green, and yellow_red
            # These are views that have the relateded unmatched events unioned to them
            if self.incremental_run and aggregation['name'] in ['timeline', 'arrival_on_green', 'yellow_red', 'split_failures']:
                aggregation['params']['incremental_run'] = True #lets the aggregation know to save unmatched events for next run
                if self.unmatched_found:
                    aggregation['params']['unmatched'] = True #lets the aggregation know to use the unmatched events from previous run
                    # split_failures uses its own view
                    if aggregation['name'] != 'split_failures':
                        aggregation['params']['from_table'] = 'raw_data_all'
              
            #######################
            ### Run Aggregation ###
            self.sql_queries[aggregation['name']] = aggregate_data(
                self.conn,
                aggregation['name'],
                self.to_sql,
                **aggregation['params'])

            end_time = time.time()
            self.runtimes[aggregation['name']] = end_time - start_time

        v_print(f"\n\nTotal aggregation runtime: {sum(self.runtimes.values()):.2f} seconds.", self.verbose)
        v_print("\nIndividual Query Runtimes:", self.verbose)
        for name, runtime in self.runtimes.items():
            v_print(f"{name}: {runtime:.2f} seconds", self.verbose)
    
    def save(self):
        """Saves the processed data."""
        if self.to_sql:
            v_print("to_sql option is True, data will not be saved.", self.verbose)
            return
        save_data(**self.__dict__)
        
    def close(self):
        """Closes the database connection."""
        self.conn.close()

    def run(self):
        """Runs the complete data processing pipeline."""
        self.load()
        self.aggregate()
        if self.to_sql:
            return self.sql_queries
        self.save()
        self.close()